#ifndef _MVBLUEFOX3_H
#define _MVBLUEFOX3_H

#include "mvbluefox3_mvfunc.h"
#include "mutex/mutex.h"
#include "mvbluefox3_settings.h"

#include <mvIMPACT_CPP/mvIMPACT_acquire.h>
#include <mvIMPACT_CPP/mvIMPACT_acquire_GenICam.h>

#include <sys/stat.h>

// Forward delcaration
namespace cv { class Mat; }

namespace CMvbluefox3 {

/** 
 * \brief Camera parameters
 *
 * Camera parameters.
 */
class CParams
{
private:

  void init(void)
  { 
    // Image format
    pixel_format = mono8;
    width = 1280;
    height = 960;
    img_rot_angle = 0.0;
    mirror = 0;
    set_aoi = false;
    aoi_width = 1280;
    aoi_height = 960;
    aoi_start_x = 0;
    aoi_start_y = 0;
    h_binning_enbl = false;
    h_binning = 1;
    v_binning_enbl = false;
    v_binning = 1;

    // Acquisition 
    frame_rate  = 100000;
    pixel_clock = 66000;
    req_timeout = 2000;

    auto_ctrl_enbl = false;
    auto_ctrl_speed = 1;
    auto_expose = true;
    auto_expose_upperlimit = 32759;
    auto_expose_des_grey_val = 50;
    expose_us = 10000;
    auto_gain = false;
    gain_db = 0;
    gamma = false;    

    // Image processing
    whiteblnce_auto_enbl = true;
    wb_r_gain = 1.0;
    wb_b_gain = 1.0;
    auto_blck_level = true;
    blck_level = 0;
    hdr_enbl = false;
    dark_cfilter = 0;
    twist_cfilter = false;
  }

public:

  codings_t pixel_format;       // Pixel format.
  int width;                    // Image width (pixels).
  int height;                   // Image height (pixels).
  float img_rot_angle;          // Image rotation angle.
  int mirror;                   // Mirror image.
  bool set_aoi;                 // Set or not AOI. 
  int aoi_width;                // AOI width.
  int aoi_height;               // AOI height.
  int aoi_start_x;              // AOI start x pixel.
  int aoi_start_y;              // AOI start y pixel.
  bool h_binning_enbl;          // Horizontal binning (enable).
  int h_binning;                // Horizontal binning factor.
  bool v_binning_enbl;          // Vertical binning (enable).
  int v_binning;                // Vertical binning factor.

  int frame_rate;               // Frame rate.
  int pixel_clock;              // Pixel_clock.
  int req_timeout;              // request timeout [ms].

  bool auto_ctrl_enbl;          // Auto controller mode (enable).
  int auto_ctrl_speed;          // Speed of auto controller mode.
  bool auto_expose;             // Exposure mode.
  int auto_expose_upperlimit;   // Auto expoure upper limit [us]. 
  int auto_expose_des_grey_val; // Desired grey value (auto controller) [%].
  int expose_us;                // Exposure time [us].
  bool auto_gain;               // Brightness auto gain (enable). 
  float gain_db;                // Brightness gain DB. 
  bool gamma;                   // Gamma.
  
  bool whiteblnce_auto_enbl;    // Auto white balance enable.
  float wb_r_gain;              // White balance red gain.
  float wb_b_gain;              // White balance blue gain.
  bool auto_blck_level;         // Auto Black level.
  int blck_level;               // Black level (if auto_blck_level = false).
  bool hdr_enbl;                // High dynamic range (enable).
  int dark_cfilter;             // Dark current filter.
  bool twist_cfilter;           // Color twist filter.

  /**
   * \brief Class constructor
   *
   * Class constructor.
   */
  CParams(){init();};

  /**
   * \brief Class destructor
   *
   * Class destructor.
   */  
  ~CParams(){};
  
};

class CMvbluefox3
{
  private:

    CParams params_; // Image parameters.

    mvIMPACT::acquire::DeviceManager devMgr_;            // Device Manager.
    mvIMPACT::acquire::Device *device_;                  // Device.
    mvIMPACT::acquire::FunctionInterface *fi_;           // Function interface.
    mvIMPACT::acquire::Request *request_;                // Request.
    mvIMPACT::acquire::SettingsBlueFOX *bf_sets_;              // BlueFox.
    mvIMPACT::acquire::GenICam::AcquisitionControl *ac_ctrl_;  // Acquisition.
    mvIMPACT::acquire::GenICam::ImageFormatControl *img_ctrl_; // Image.
    mvIMPACT::acquire::GenICam::AnalogControl *anlg_ctrl_;     // Analog.
    mvIMPACT::acquire::GenICam::DeviceControl *dev_ctrl_;      // Device.
    mvIMPACT::acquire::GenICam::LUTControl *lut_ctrl_;         // LUT.
    mvIMPACT::acquire::Statistics *stats_;                     // Statistics. 

    int timeout_ms_;    // Frame request timeout.

    codings_t coding_;  // Image coding.
    int linepitch_;     // Line pitch.    
    int depth_;           // Bits per image channel.
    int bytes_per_pixel_; // Bytes per pixel.
    bayer_filter_t bayer_filter_;    // Pixel Color filter.

    /**
     * \brief Mutex object to handle the access to the image
     *
     * This mutex is used to avoid multiple threads to access the internal 
     * image data buffer, and therefore avoid data corruption. The Mutex
     * is automatically initialized at construction time.
     */ 
    CMutex image_access_;

    /**
     * \brief Delete pointer
     *
     *  Function to delete pointers.
     */
    template <typename Ptr_t>
    void DelPtr(Ptr_t &ptr) 
    {
      if (ptr != NULL) 
      {
        ptr = NULL;
        delete [] ptr;
      }
    }

    /**
     * \brief Open device
     *
     *  Open device.
     */
    void OpenDevice(const std::string &sett_path = std::string());

    /**
     * \brief Get image request format
     *
     *  Fill all image request format variables by quering the request.
     */
    void GetImageReqFormat(mvIMPACT::acquire::Request *request);

    /**
     * \brief Request single image
     *
     *  Sends a single request to the default request queue of the device.
     */
    void RequestSingle(void);

    /**
     * \brief Request multiple images
     *
     *  Sends multiple (n) requests to the default request queue of the device.
     */
    void RequestMultiple(int n);

    /**
     * \brief Pixel clock to frame rate
     *
     *  Convert pixel clock to frame rate.
     * http://www.matrix-vision.com/manuals/mvBlueFOX/
     *        Appendix_page_0.html#CMOS752_section_1_1
     */
    int PixelClockToFrameRate(int &pclk_khz, int &width, int &height, int &expose_us); 

    /**
     * \brief Set image size
     *
     *  Set image size.
     */    
    void SetImgSize(int &width, int &height);

    /**
     * \brief Set pixel format.
     *
     *  Set pixel format.
     */  
    void SetPixelFormat(codings_t &pf); 

    /**
     * \brief Rotate Image.
     *
     *  Rotate Image (Disabled with with rot_angle = 0).
     */ 
    void RotateImg(float &rot_angle);

    /**
     * \brief Mirror Image.
     *
     *  Mirror Image.
     */ 
    void SetImgMirror(int type);

    /**
     * \brief Set aera of intereset.
     *
     *  Set aera of intereset.
     */   
    void SetAoi(int &width, int &height, int &start_x, int &start_y); 

    /**
     * \brief Set image horizontal binning.
     *
     *  Set image horizontal binning.
     */ 
    void SetHBinning(bool &h_bin_enbl, int &h_bin);

    /**
     * \brief Set image vertical binning.
     *
     *  Set image vertical binning.
     */ 
    void SetVBinning(bool &v_bin_enbl, int &v_bin);    
  
    /**
     * \brief Set Auto Gain Control.
     *
     *  Set Auto Gain Control (analogic). Gain [dB] is set in case of manual gain 
     *  (auto_gain = false).
     */ 
    void SetAutoGainCtrl(bool &auto_gain, float &gain_db);

    /**
     * \brief Set Expose Control.
     *
     *  Set Auto Expose Control. Exposure [us] is set in case of manual exposure (auto_expose = false).
     */ 
    void SetAutoExposCtrl(bool &auto_expose, int &auto_upperlimit, int &des_gray_val, int &expose_us);

    /**
     * \brief Sets Auto Control Speed.
     *
     *  Sets Auto Control Speed and desired average grey value.
     */ 
    void SetAutoCtrl(bool &actrl_enbl, int &auto_ctrl_speedpeed, int &des_gray_val);

    /**
     * \brief Sets white balance.
     *
     *  Sets white balance.
     */ 
    void SetWhiteBlnce(bool &wbp, float &r_gain, float &b_gain);

    /**
     * \brief Sets High Dynamic Range.
     *
     *  Sets High Dynamic Range.
     */ 
    void SetHdr(bool &hdr);

    /**
     * \brief Sets Dark Current Filter.
     *
     *  Sets Dark Current Filter.
     */ 
    void SetDarkCFilter(int &dcfm);

    /**
     * \brief Set Frame Rate 
     *
     *  Set Frame Rate [Hz]. It is subject to other values like pixel clock 
     *  or exposure. 
     */ 
    void SetFrameRate(int &cpc); 

    /**
     * \brief Set Pixel Clock.
     *
     *  Set Pixel Clock [kHz].
     */ 
    void SetPixelClock(int &pclk_khz); 

    /**
     * \brief Set timeout [ms]
     *
     *  Sets the timeout for the frame request operation.
     */
    void SetTimeoutMS(int &timeout_ms);

    /**
     * \brief Get timeout [ms]
     *
     *  Returns the timeout set in the frame request operation.
     */
    int GetTimeoutMS(void);

    /**
     * \brief Set Gamma.
     *
     *  Set Gamma (LUT operations).
     */     
    void SetGamma(bool &enbl);

    /**
     * \brief Set Black Level.
     *
     *  Set Black Level.
     */  
    void SetBlackLevel(bool &auto_blck_level, int &blck_level);

    /**
     * \brief Color twsit filter.
     *
     *  Color twsit filter.
     */ 
    void ColorTwistFilter(bool &ct_enbl);

    /**
     * \brief Check if File exist.
     *
     *  Check if File exist.
     */ 
    bool FileExist(const std::string& name);

  public:

    /**
     * \brief Class constructor
     *
     * Class constructor.
     */
    CMvbluefox3(const mvIMPACT::acquire::Device *device, const std::string &sett_path = std::string());

    /**
     * \brief Class destructor
     *
     * Class destructor.
     */    
    ~CMvbluefox3();

    /**
     * \brief Close device
     *
     *  Close device.
     */
    void CloseDevice(void);

    /**
     * \brief Load camera settings (previously stored with wxPropView
     *
     *  Load camera settings (previously stored with wxPropView.
     */    
    void LoadSettings(const std::string &path);
    
    /**
     * \brief Get device product name
     *
     * Get device product name.
     */   
    std::string GetProduct(void); 
    
    /**
     * \brief Get device serial number
     *
     * Get device serial number.
     */       
    std::string GetSerial(void); 
    
    /**
     * \brief Get single image
     *
     *  Sends a request to the default request queue of the device, 
     *  waits for the result and returns the image.
     */
    void GetImage(char **image);
#ifdef HAVE_OPENCV_H
    void GetImageCV(cv::Mat &image);
#endif

    /**
     * \brief Get 'real' exposure time [us]
     *
     *  Returns the 'real' expose time (in us) used to generate the image.
     */
    int ReqExposeUs(void);

    /**
     * \brief Set Configuration to the device.
     *
     *  Set main camera parameters.
     */
    void SetConfig(CParams &params, bool ini = false); 

    /**
     * \brief Get Configuration.
     *
     *  Get main camera parameters.
     */    
    CParams GetConfig(void); 

    /**
     * \brief Get bytes per pixel 
     *
     *  Get bytes per pixel.
     */ 
    int GetBytesPixel(void);

    /**
     * \brief Get image depth (channels)
     *
     *  Get image depth (channels).
     */ 
    int GetDepth(void);

    /**
     * \brief Get line pitch 
     *
     *  Get line pitch.
     */ 
    int GetImgLinePitch(void);

    /**
     * \brief Get bayer color filter format
     *
     *  Get bayer color filter format.
     */
    bayer_filter_t GetBayerColorFilter(void);
};

} // End of CMvbluefox3 namespace

#endif
