#ifndef _MVBLUEFOX3_SETTINGS_H
#define _MVBLUEFOX3_SETTINGS_H

#include "exceptions/mvbluefox3_exceptions.h"

#include <mvIMPACT_CPP/mvIMPACT_acquire.h>

#include <iostream>
#include <algorithm>
#include <boost/type_traits/is_enum.hpp>

namespace CMvbluefox3 {

/** 
 * \brief Identifiers of the pixel format codings
 *
 * Enumeraion of all possible pixel format codings supported by the cameras.
 * The following macros allow to create the enum codings_t but also their string format codings_str. Its usage to get the string is:
 * codings_str[codings_t].
 */
#define FOREACH_ENCODING(ENCODING) \
  ENCODING(raw) \
  ENCODING(mono8) \
  ENCODING(mono10) \
  ENCODING(mono12) \
  ENCODING(mono12v1) \
  ENCODING(mono12v2) \
  ENCODING(mono14) \
  ENCODING(mono16) \
  ENCODING(bgr8) \
  ENCODING(bgr8p) \
  ENCODING(bgr10) \
  ENCODING(bgr12) \
  ENCODING(bgr14) \
  ENCODING(bgr16) \
  ENCODING(bgra8) \
  ENCODING(bgra8p) \
  ENCODING(rgb8) \
  ENCODING(rgb10v2) \
  ENCODING(yuv411uyyvyy) \
  ENCODING(yuv422) \
  ENCODING(yuv422p) \
  ENCODING(yuv42210) \
  ENCODING(yuv422uyvy) \
  ENCODING(yuv422uyvy10) \
  ENCODING(yuv444) \
  ENCODING(yuv44410p) \
  ENCODING(yuv444uyv) \
  ENCODING(yuv444uyv10) \
  ENCODING(bayer_rggb8) \
  ENCODING(bayer_gbrg8) \
  ENCODING(bayer_grbg8) \
  ENCODING(bayer_bggr8) \
  ENCODING(bayer_rggb16) \
  ENCODING(bayer_gbrg16) \
  ENCODING(bayer_grbg16) \
  ENCODING(bayer_bggr16) 
#define GENERATE_ENUM_ENC(ENUM) ENUM,
#define GENERATE_STRING_ENC(STRING) #STRING,
enum codings_t { FOREACH_ENCODING(GENERATE_ENUM_ENC) };
static const char *codings_str[] = { FOREACH_ENCODING(GENERATE_STRING_ENC) };
inline void empty_funca(void){(void)codings_str;}

/** 
 * \brief Identifiers of the binning codings
 *
 * Enumeraion of all possible binning codings supported by the cameras.
 * The following macros allow to create the enum binnings_t but also their string format binnings_str. Its usage to get the string is:
 * binnings_str[binnings_t].
 */
#define FOREACH_BINNING(BINNING) \
  BINNING(Bin_Off) \
  BINNING(Bin_H) \
  BINNING(Bin_V) \
  BINNING(Bin_HV) \
  BINNING(Bin_3H) \
  BINNING(Bin_3V) \
  BINNING(Bin_3H3V) \
  BINNING(Bin_HAvg) \
  BINNING(Bin_VAvg) \
  BINNING(Bin_HVAvg) \
  BINNING(Bin_3HAvg) \
  BINNING(Bin_3VAvg) \
  BINNING(Bin_3H3VAvg) \
  BINNING(Drop_H) \
  BINNING(Drop_V) \
  BINNING(Drop_HV) \
  BINNING(Drop_3H) \
  BINNING(Drop_3V) \
  BINNING(Drop_3H3V)
#define GENERATE_ENUM_BIN(ENUM) ENUM,
#define GENERATE_STRING_BIN(STRING) #STRING,
enum binnings_t { FOREACH_BINNING(GENERATE_ENUM_BIN) };
static const char *binnings_str[] = { FOREACH_BINNING(GENERATE_STRING_BIN) };
inline void empty_funcb(void){(void)binnings_str;}

#define FOREACH_BAYERFILTER(COLORFILTER) \
  COLORFILTER(none) \
  COLORFILTER(bayer_rg) \
  COLORFILTER(bayer_gb) \
  COLORFILTER(bayer_gr) \
  COLORFILTER(bayer_bg)
#define GENERATE_ENUM_BF(ENUM) ENUM,
#define GENERATE_STRING_BF(STRING) #STRING,
enum bayer_filter_t { FOREACH_BAYERFILTER(GENERATE_ENUM_BF) };
static const char *bayer_filter_str[] = { FOREACH_BAYERFILTER(GENERATE_STRING_BF) };
inline void empty_funcc(void){(void)bayer_filter_str;}

/** 
 * \brief Matrix Vision Pixel format to Encoding
 *
 * Matrix Vision Pixel format to Encoding.
 */
codings_t MVFormatToEncoding(const TImageDestinationPixelFormat& pixel_format); 

/** 
 * \brief Encoding to Matrix Vision Pixel format
 *
 * Codings to Matrix Vision Pixel format.
 */
TImageDestinationPixelFormat EncodingToMVFormat(const codings_t& coding); 

/** 
 * \brief Matrix Vision Bayer Pattern to Encoding 
 *
 * Matrix Vision Bayer Pattern to Encoding.
 */
codings_t BayerPatternToEncoding(const TBayerMosaicParity& bayer_pattern,
                                  int bytes_per_pixel);

/** 
 * \brief Matrix Vision Pixel Color Format to bayer_filter_t
 *
 * Matrix Vision Pixel Color Format to bayer_filter_t.
 */

bayer_filter_t MVFormatToColorFilter(const int &pcf);

/** 
 * \brief Matrix Vision TCameraBinningMode to binning_t
 *
 * Matrix Vision TCameraBinningMode to binning_t.
 */
binnings_t MVBinningToBinning(const TCameraBinningMode& idbm);

/** 
 * \brief binning_t to Matrix Vision TCameraBinningMode 
 *
 * binning_t to Matrix Vision TCameraBinningMode.
 */
TCameraBinningMode BinningToMVBinning(const binnings_t& bin);

/** 
 * \brief TranslationDict typedef
 *
 * TranslationDict typedef.
 */
template<typename ValueType>
struct TranslationDict { typedef std::vector<std::pair<std::string, ValueType> > type;
};

/** 
 * \brief Function template to compare values (clamp)
 *
 * If value compares less than high, returns the larger of value and low, otherwise returns the smaller of value and high. 
 */
template <typename T1, typename T2>
T1 Clamp(const T1& value, const T2& low, const T2& high) 
{
  return std::max<T1>(low, std::min<T1>(high, value));
}

/** 
 * \brief Function template to compare properties (clamp)
 *
 * It compares the property by using the function template clamp.
 */
template <typename PropertyType, typename ValueType>
ValueType ClampProperty(const PropertyType& prop, const ValueType& value) 
{
  return Clamp(value, prop.getMinValue(), prop.getMaxValue());
}

/**
 * @brief GetTranslationDict Convenience function to get translation dict
 * @param prop Property
 * @return Translation dict
 */
template <typename PropertyType>
typename TranslationDict<typename PropertyType::value_type>::type GetTranslationDict(const PropertyType& prop) 
{
  typename TranslationDict<typename PropertyType::value_type>::type dict;
  prop.getTranslationDict(dict);
  return dict;
}

/** 
 * \brief Print property
 *
 * Print property values.
 */
template <typename Prop>
void PrintProperty(const Prop& prop) {
  std::cout << "Property name: " << prop.displayName() << "\n";
  if (prop.hasMaxValue()) 
    std::cout << "  Max value: " << prop.getMaxValue() << "\n";
  if (prop.hasMinValue()) 
    std::cout << "  Min value: " << prop.getMinValue() << "\n";
  if (prop.hasStepWidth()) 
    std::cout << "  Step width: " << prop.getStepWidth() << "\n";
  std::cout << "  Writable: " << prop.isWriteable() << "\n";
  std::cout << "  Visible: " << prop.isVisible() << "\n";
  std::cout << "  Valid: " << prop.isValid() << std::endl;
}

/** 
 * \brief Read property
 *
 * Read property values.
 */
template <typename PropertyType, typename ValueType>
void ReadProperty(const PropertyType& prop, ValueType& value) 
{
  if (!prop.isValid()) 
    throw CmvBlueFOX3Exception(_HERE_, "Unable to read property: " + prop.name() + std::string(". Property not valid."));  

  try {
    value = static_cast<ValueType>(prop.read());
  } catch (mvIMPACT::acquire::ImpactAcquireException& e ) {
  	throw CmvBlueFOX3Exception(_HERE_, "Unable to read property: " + prop.name() + std::string(" (error code: ") + std::string(e.getErrorCode() + ").") );
  }
}

/** 
 * \brief Write property
 *
 * Write property values.
 */
template <typename PropertyType, typename ValueType>
void WriteProperty(const PropertyType& prop, ValueType value) 
{
  // Check if it's possible to write to this property
  if (!prop.isVisible()) 
    throw CmvBlueFOX3Exception(_HERE_, "Unable to write property: " + prop.name() + std::string(". Property not visible."));
  if (!prop.isWriteable()) 
    throw CmvBlueFOX3Exception(_HERE_, "Unable to write property: " + prop.name() + std::string(". Property not not writeable."));
  if (!prop.isValid()) 
    throw CmvBlueFOX3Exception(_HERE_, "Unable to write property: " + prop.name() + std::string(". Property not valid."));

  // Clamp value to valid range only if it's not an enum
  if (!boost::is_enum<typename PropertyType::value_type>::value)    
    if (prop.hasMaxValue() && prop.hasMinValue()) 
      value = ClampProperty(prop, value);

  try {
    prop.write(static_cast<typename PropertyType::value_type> (value));
  } catch (mvIMPACT::acquire::ImpactAcquireException& e ) {
  	throw CmvBlueFOX3Exception(_HERE_, "Unable to write property: " + prop.name() + std::string(" (error code: ") + std::string(e.getErrorCode() + ").") );
  }
}

/** 
 * \brief Write and read property
 *
 * Write and read property values.
 */
template <typename PropertyType, typename ValueType>
void WriteAndReadProperty(const PropertyType& prop, ValueType& value) 
{
  WriteProperty(prop, value);
  ReadProperty(prop, value);
}

} // End of CMvbluefox3 namespace

#endif