#include "mvbluefox3.h"
#include <unistd.h>
#include <limits.h>
#include <sstream>

#ifdef HAVE_OPENCV_H
#include <opencv2/core/version.hpp>
#if CV_MAJOR_VERSION == 3
#include <opencv2/opencv.hpp>
#endif
#endif

typedef bool( *SUPPORTED_DEVICE_CHECK )( const mvIMPACT::acquire::Device* const );

// mvIMPACT::acquire::DeviceManager devMgr_;            // Device Manager.


std::string GetSerialFromUser(void)
{
  mvIMPACT::acquire::DeviceManager devMgr; // Device Manager.
  return getDeviceFromUserInput(devMgr)->serial.read();
}

std::string AvailableDevices(SUPPORTED_DEVICE_CHECK pSupportedDeviceCheckFn) 
{
  mvIMPACT::acquire::DeviceManager devMgr; // Device Manager.
  std::ostringstream devices;
  devices << "Available device(s): ";
  const unsigned int devCnt = devMgr.deviceCount();
  if (devCnt == 0)
  {
    devices << "0.";
  }
  else
  {
    for( unsigned int i = 0; i < devCnt; i++ )
    {
      Device* pDev = devMgr[i];
      if( pDev )
      {
        if( !pSupportedDeviceCheckFn || pSupportedDeviceCheckFn( pDev ) )
        {
          devices << " \n  - " << pDev->product.read() << ": " << pDev->serial.read();
        }
      }
    }
  }
  return devices.str();
}

std::string GetPath()
{
  char result[ PATH_MAX ];
  ssize_t count = readlink( "/proc/self/exe", result, PATH_MAX );
  return std::string( result, (count > 0) ? count : 0 );
}

int main(int argc, char *argv[])
{
  int num_cams = 1;

  // Get own path
  std::string argv_str(argv[0]);
  std::string pbase = argv_str.substr(0, argv_str.find_last_of("/"));
  std::string paramspath = pbase + "/../src/examples/example_params.xml";

  std::cout << std::endl;
  std::cout << "******************************" << std::endl;
  std::cout << "* ----  mvBlueFOX3 test ---- *" << std::endl;
  std::cout << "******************************" << std::endl;
  std::cout << std::endl;
  std::cout << "Please, choose the device to test from the following list:" << std::endl;
  std::cout << std::endl;

  std::vector<std::string> serial;
  serial.resize(num_cams);
  serial[0] = GetSerialFromUser();

#if defined(HAVE_OPENCV_H) && CV_MAJOR_VERSION == 3
  std::vector<std::string> window_name;
  window_name.resize(num_cams);
  window_name[0] = serial[0];
#endif

  mvIMPACT::acquire::DeviceManager devMgr_;       // Device Manager.
  std::vector<CMvbluefox3::CMvbluefox3*> cam_ptr; // Device.

  try {
  
    cam_ptr.resize(num_cams);
    for (int ii = 0; ii < num_cams; ++ii)
    {
      cam_ptr[ii] = NULL;

      if( !serial[ii].empty() )
      {
        std::cout << "Trying to open device with serial: " << serial[ii] << std::endl;

        if (devMgr_.getDeviceBySerial(serial[ii]) == 0)
          std::cout << "No device found! Unable to continue. " << std::endl;
      else
        cam_ptr[ii] = new CMvbluefox3::CMvbluefox3(devMgr_.getDeviceBySerial(serial[ii]),paramspath);
      }
      else
        std::cout << "Device cannot be found because serial number is not specified." << std::endl;
    }
  
#if defined(HAVE_OPENCV_H) && CV_MAJOR_VERSION == 3

    std::cout << "Acquiring images." << std::endl << ">WARNING: depending on OpenCV compilation flags, visualization might experience hard delays (e.g. if \"init done\" appears in the following line)" << std::endl;

    for (int ii = 0; ii < num_cams; ++ii)
      cv::namedWindow( window_name[ii], cv::WINDOW_NORMAL );

    while (true) 
    {
      for (int ii = 0; ii < num_cams; ++ii)
      {
        cv::Mat image;
        cam_ptr[ii]->GetImageCV(image);
        cv::imshow( window_name[ii], image );
      }
      if(cv::waitKey(30) >= 0) break;
    }

    cv::destroyAllWindows();    

#else

    for (int ii = 0; ii < num_cams; ++ii)
    {
      char *image=NULL;
      cam_ptr[ii]->GetImage(&image);
      std::cout << "[Camera test]: CAM: " << serial[ii] << ". Frame acquired successfully." << "OpenCV 3 is required and not installed, thus visualization is avoided." << std::endl;
      delete [] image;
    }

#endif

    std::cout << "[Camera test]: Finished successfully." << std::endl;
  }catch (CMvbluefox3::CmvBlueFOX3Exception& e) {
    std::cout << e.what() << std::endl;
  }
  
  for (int ii = 0; ii < num_cams; ++ii)
  {
    if (cam_ptr[ii] != NULL) 
    {
      cam_ptr[ii] = NULL;
      delete [] cam_ptr[ii];
    }
  }
}
