#include "mvbluefox3.h"
#include "exceptions/mvbluefox3_exceptions.h"

#include <iostream>
#include <sstream>

#ifdef HAVE_OPENCV_H
#include <opencv2/opencv.hpp>
#endif

// comment this line to see the effect of queue underruns
#define DROP_OUTDATED_IMAGES
// uncomment this line to get more output into the console window
//#define VERBOSE_OUTPUT

// mvIMPACT::acquire::DeviceManager devMgr_;            // Device Manager.

namespace CMvbluefox3 {

CMvbluefox3::CMvbluefox3(const mvIMPACT::acquire::Device *device, const std::string &sett_path)
{
  // Initialize vars
  this->device_ = new mvIMPACT::acquire::Device(*device);
  this->timeout_ms_ = -1;
  this->linepitch_ = 8;
  this->depth_ = 8;
  this->bytes_per_pixel_ = 1;

  OpenDevice(sett_path);
}

CMvbluefox3::~CMvbluefox3()
{
  // Close device if necessary.  
  CloseDevice();
  
  // Delete all pointers if needed.
  DelPtr(this->device_);
  DelPtr(this->fi_);
  DelPtr(this->request_);
  DelPtr(this->bf_sets_);
  DelPtr(this->ac_ctrl_);
  DelPtr(this->img_ctrl_);
  DelPtr(this->anlg_ctrl_);
  DelPtr(this->dev_ctrl_);
  DelPtr(this->lut_ctrl_);
  DelPtr(this->stats_);
}

void CMvbluefox3::OpenDevice(const std::string &sett_path)
{
  // Open the device
  try { this->device_->open(); }
  catch(mvIMPACT::acquire::ImpactAcquireException& e )
  {
    throw CmvBlueFOX3Exception(_HERE_, "An error occurred while opening the device (error code: " + e.getErrorCode() + std::string(").") );
  }

  // These poniters will probably leak, but we don't really care
  this->fi_ = new mvIMPACT::acquire::FunctionInterface(this->device_);
  this->bf_sets_ = new mvIMPACT::acquire::SettingsBlueFOX(this->device_);
  this->ac_ctrl_ = new mvIMPACT::acquire::GenICam::AcquisitionControl(this->device_);
  this->img_ctrl_ = new mvIMPACT::acquire::GenICam::ImageFormatControl(this->device_);
  this->anlg_ctrl_ = new mvIMPACT::acquire::GenICam::AnalogControl(this->device_);
  this->dev_ctrl_ = new mvIMPACT::acquire::GenICam::DeviceControl(this->device_);
  this->lut_ctrl_ = new mvIMPACT::acquire::GenICam::LUTControl(this->device_);
  this->stats_ = new mvIMPACT::acquire::Statistics(this->device_);

  this->bf_sets_->cameraSetting.restoreDefault();
  this->bf_sets_->imageDestination.restoreDefault();
  this->bf_sets_->imageProcessing.restoreDefault();

  if (!sett_path.empty())
    LoadSettings(sett_path);
  else
    SetConfig(this->params_,true);
}

void CMvbluefox3::CloseDevice(void)
{
  if (this->device_ && this->device_->isOpen()) 
  {
    std::cout << "[mvBlueFOX3]: Closing device with serial: " << this->device_->serial.readS() << std::endl;
    
    // free resources, empty the drivers request queue
    if (this->fi_!=NULL)
      this->fi_->imageRequestReset( 0, 0 );

    this->device_->close();
  }
}

void CMvbluefox3::LoadSettings(const std::string &path)
{
  if (FileExist(path))
  {  
    std::cout << "[mvBlueFOX3]: Loading stored settings for " << this->device_->product.readS() << " " << this->device_->serial.readS() << " from path: " << path << "." << std::endl;
    this->fi_->loadSetting(path,TStorageFlag(sfFile));
  }
  else
    throw CmvBlueFOX3Exception(_HERE_,"Parameters cannot be loaded because parameters file (XML) does not exist. Check XML file path;" + path);
}

bool CMvbluefox3::FileExist(const std::string& name) 
{
  struct stat buffer;   
  return (stat (name.c_str(), &buffer) == 0); 
}

std::string CMvbluefox3::GetProduct(void) 
{ 
  if (this->device_==NULL)
    throw CmvBlueFOX3Exception(_HERE_,"Product cannot be retrieved. Device not found.");
  return this->device_->product.readS(); 
}

std::string CMvbluefox3::GetSerial(void) 
{ 
  if (this->device_==NULL)
    throw CmvBlueFOX3Exception(_HERE_,"Serial cannot be retrieved. Device not found.");
  return this->device_->serial.readS(); 
}

void CMvbluefox3::RequestSingle(void) 
{
  int result = DMR_NO_ERROR;
  result = this->fi_->imageRequestSingle();
  if( result != DMR_NO_ERROR )
    throw CmvBlueFOX3Exception(_HERE_, "Error while filling the request queue: (" + ImpactAcquireException::getErrorCodeAsString(result) + std::string(").") );  
}

void CMvbluefox3::RequestMultiple(int n) 
{
  for (int ii = 0; ii < n; ++ii) 
  {
    RequestSingle();
    int requestNr = this->fi_->imageRequestWaitFor(timeout_ms_);
    this->fi_->imageRequestUnlock(requestNr);
  }
}

void CMvbluefox3::GetImage(char **image)  
{ 
  // send a request to the default request queue of the device and wait for the result.
  RequestSingle();

  // Start the acquisition manually if this was requested (this is to prepare the driver for data capture and tell the device to start streaming data)
  if( this->device_->acquisitionStartStopBehaviour.read() == assbUser )
  {
    TDMR_ERROR result;
    if( (result = static_cast<TDMR_ERROR>(this->fi_->acquisitionStart()) ) != DMR_NO_ERROR )
      throw CmvBlueFOX3Exception(_HERE_, "'FunctionInterface.acquisitionStart' returned with an unexpected result: (" + ImpactAcquireException::getErrorCodeAsString(result) + std::string(").") );
  }

  // wait for results from the default capture queue
  int requestNr = this->fi_->imageRequestWaitFor( GetTimeoutMS() );

  // check if the image has been captured without any problems
  if( !this->fi_->isRequestNrValid( requestNr ) )
  {
    // If the error code is -2119(DEV_WAIT_FOR_REQUEST_FAILED), the documentation will provide
    // additional information under TDMR_ERROR in the interface reference
    throw CmvBlueFOX3Exception(_HERE_, "'imageRequestWaitFor' failed (" + ImpactAcquireException::getErrorCodeAsString(requestNr) + std::string("). Possible cause: timeout value too small?") );
  }

  this->request_ = this->fi_->getRequest(requestNr);
  if( !this->request_->isOK() )
  {
    // unlock the buffer to let the driver know that you no longer need this buffer
    this->fi_->imageRequestUnlock(requestNr);
    throw CmvBlueFOX3Exception(_HERE_, "Request result: " + this->request_->requestResult.readS() );
  }

  // Lock image access
  this->image_access_.enter();

  // Get image format
  GetImageReqFormat(this->request_);

  // Get the image from the buffer.
  *image = new char[this->request_->imageSize.read()];
  memcpy( *image, this->request_->imageData.read(), this->request_->imageSize.read() );
 
  // Unlock image access
  this->image_access_.exit();

  // unlock the buffer to let the driver know that you no longer need this buffer
  this->fi_->imageRequestUnlock(requestNr);

  // DEBUG
  // std::cout << this->stats_->framesPerSecond.read() << std::endl;
  // std::cout << this->ac_ctrl_->mvResultingFrameRate.read() << std::endl;
}

#ifdef HAVE_OPENCV_H
void CMvbluefox3::GetImageCV(cv::Mat &image)
{
  char *image_raw=NULL;
  GetImage(&image_raw);
  
  int type = CV_8UC1;

  switch (this->depth_*this->bytes_per_pixel_) 
  {
    case 8: type = CV_8UC1; break;
    case 16: type = CV_8UC2; break;
    case 24: type = CV_8UC3; break;
    case 32: type = CV_8UC4; break;
    default: throw CmvBlueFOX3Exception(_HERE_,
      "Invalid depth. OpenCV Depth not implemented here."); break;
  }

  image = cv::Mat(this->params_.height,this->params_.width,type,image_raw);   
  cv::cvtColor(image, image, CV_BGR2RGB); // Opencv uses BGR.

  DelPtr(image_raw);
}
#endif

void CMvbluefox3::GetImageReqFormat(mvIMPACT::acquire::Request *request)
{
  if ( request == NULL)
    throw CmvBlueFOX3Exception(_HERE_, "Invalid request reference.");

  // Image width ImageDestination   
  this->params_.height = this->request_->imageHeight.read();   

  // Image height ImageDestination   
  this->params_.width = this->request_->imageWidth.read();

  // Channel bit depth
  ReadProperty(this->request_->imageChannelBitDepth,this->depth_);

  // Bytes per pixel
  ReadProperty(this->request_->imageBytesPerPixel,this->bytes_per_pixel_);

  // Line pitch
  ReadProperty(this->request_->imageLinePitch,this->linepitch_);

  // Color filter
  int pcolfilter;
  ReadProperty(this->img_ctrl_->pixelColorFilter,pcolfilter);
  this->bayer_filter_ = MVFormatToColorFilter(pcolfilter);

  // Bayer parity of this image if this buffer is part of a larger image
  TBayerMosaicParity bayer_mosaic_parity;
  ReadProperty(request_->imageBayerMosaicParity,bayer_mosaic_parity);
  if (bayer_mosaic_parity == bmpUndefined) 
  {
    TImageDestinationPixelFormat idpf;
    ReadProperty(this->bf_sets_->imageDestination.pixelFormat,idpf);
    this->params_.pixel_format = MVFormatToEncoding(idpf); 
  }
  else // Bayer pattern
  {
    this->params_.pixel_format = BayerPatternToEncoding(bayer_mosaic_parity,
                                                      this->bytes_per_pixel_);
  }
}

int CMvbluefox3::GetBytesPixel(void)
{
  return this->bytes_per_pixel_;
}

int CMvbluefox3::GetDepth(void)
{
  return this->depth_;
}

int CMvbluefox3::GetImgLinePitch(void)
{
  return this->linepitch_;
}

bayer_filter_t CMvbluefox3::GetBayerColorFilter(void)
{
  return this->bayer_filter_;
}

int CMvbluefox3::ReqExposeUs(void) 
{
  if (this->request_ == NULL || !this->request_->isOK()) 
    throw CmvBlueFOX3Exception(_HERE_, "An error occurred while retrieveing exposure time. Request not found.");

  return this->request_->infoExposeTime_us.read();
}

void CMvbluefox3::SetConfig(CParams &params, bool ini) 
{
  // Clear request queue
  this->fi_->imageRequestReset(0, 0);

  // Image size
  if (ini || this->params_.width != params.width || this->params_.height != params.height)
    SetImgSize(params.width,params.height);

  // Pixel Format
  if (ini || this->params_.pixel_format != params.pixel_format)
    SetPixelFormat(params.pixel_format);

  // Image Rotation
  if (ini || this->params_.img_rot_angle != params.img_rot_angle)
    RotateImg(params.img_rot_angle);

  // Image mirror
  if (ini || this->params_.mirror != params.mirror)
    SetImgMirror(params.mirror);

  // Area of Intreset
  if (ini || this->params_.aoi_width != params.aoi_width || this->params_.aoi_height != params.aoi_height || this->params_.aoi_start_x != params.aoi_start_x || this->params_.aoi_start_y != params.aoi_start_y)
    SetAoi(params.aoi_width, params.aoi_height, params.aoi_start_x, params_.aoi_start_y);

  // Binning
  if (ini || this->params_.h_binning_enbl != params.h_binning_enbl || this->params_.h_binning != params.h_binning)
    SetHBinning(params.h_binning_enbl, params.h_binning);
  if (ini || this->params_.v_binning_enbl != params.v_binning_enbl || this->params_.v_binning != params.v_binning)  
    SetVBinning(params.v_binning_enbl, params.v_binning);

  // Gain
  if (ini || this->params_.auto_gain != params.auto_gain || this->params_.gain_db != params.gain_db)  
    SetAutoGainCtrl(params.auto_gain, params.gain_db);

  // Auto Controller
  if (ini || this->params_.auto_ctrl_enbl != params.auto_ctrl_enbl || this->params_.auto_ctrl_speed != params.auto_ctrl_speed || this->params_.auto_expose_des_grey_val != params.auto_expose_des_grey_val)    
    SetAutoCtrl(params.auto_ctrl_enbl, params.auto_ctrl_speed, params.auto_expose_des_grey_val);

  // White Balance
  if (ini || this->params_.whiteblnce_auto_enbl != params.whiteblnce_auto_enbl || this->params_.wb_r_gain != params.wb_r_gain || this->params_.wb_b_gain != params.wb_b_gain)    
    SetWhiteBlnce(params.whiteblnce_auto_enbl, params.wb_r_gain, params.wb_b_gain);

  // High Dynamic Range
  if (ini || this->params_.hdr_enbl != params.hdr_enbl)
    SetHdr(params.hdr_enbl);

  // Dark Current Filter
  if (ini || this->params_.dark_cfilter != params.dark_cfilter)
    SetDarkCFilter(params.dark_cfilter);

  // Request time out
  if (ini || this->params_.req_timeout != params.req_timeout)
    SetTimeoutMS(params.req_timeout);
  
  // Pixel Clock
  if (ini || this->params_.pixel_clock != params.pixel_clock)
  {
    SetPixelClock(params.pixel_clock);
    params.frame_rate = this->params_.frame_rate;
  }

  // Frame rate
  if (ini || this->params_.frame_rate != params.frame_rate)
    SetFrameRate(params.frame_rate);

  // Exposure
  if (ini || this->params_.auto_expose != params.auto_expose || this->params_.auto_expose_upperlimit != params.auto_expose_upperlimit || this->params_.auto_expose_des_grey_val != params.auto_expose_des_grey_val || this->params_.expose_us != params.expose_us)
    SetAutoExposCtrl(params.auto_expose, params.auto_expose_upperlimit, params.auto_expose_des_grey_val, params.expose_us);

  // Gamma
  if (ini || this->params_.gamma != params.gamma)
    SetGamma(params.gamma);

  // Black Level
  if (ini || this->params_.auto_blck_level != params.auto_blck_level || this->params_.blck_level != params.blck_level)
    SetBlackLevel(params.auto_blck_level, params.blck_level);

  // Twsit color filter
  if (ini || this->params_.twist_cfilter != params.twist_cfilter)
    ColorTwistFilter(params.twist_cfilter);

  // Cache these parameters
  params_ = params;
}

CParams CMvbluefox3::GetConfig(void) 
{
  return this->params_;
}


void CMvbluefox3::SetImgSize(int &width, int &height)
{
  try{    
    WriteProperty(this->img_ctrl_->width,width);
    WriteProperty(this->img_ctrl_->height,height);
  }catch (CmvBlueFOX3Exception& e) {
    throw CmvBlueFOX3Exception(_HERE_, "Unable to set image size: " + e.what() );
  }   
}

void CMvbluefox3::SetPixelFormat(codings_t &pf)
{ 
  TImageDestinationPixelFormat idpf = EncodingToMVFormat(pf);
  try{    
    WriteProperty(this->bf_sets_->imageDestination.pixelFormat, idpf);
  }catch (CmvBlueFOX3Exception& e) {
    throw CmvBlueFOX3Exception(_HERE_, "Unable to set pixel format: " + e.what() );
  }    
}

void CMvbluefox3::RotateImg(float &rot_angle)
{
  if (rot_angle != 0.0)
  {
    if (!this->bf_sets_->imageProcessing.rotationEnable.isValid())
      throw CmvBlueFOX3Exception(_HERE_, "Image rotation option is not available for this model.");
    try{    
      WriteProperty(this->bf_sets_->imageProcessing.rotationEnable, bTrue);
      WriteProperty(this->bf_sets_->imageProcessing.rotationAngle, rot_angle);
    }catch (CmvBlueFOX3Exception& e) {
      throw CmvBlueFOX3Exception(_HERE_, "Unable to rotate image: " + e.what() );
    }      
  }
  else
  {
    if (this->bf_sets_->imageProcessing.rotationEnable.isValid())
    {
      try{    
        WriteProperty(this->bf_sets_->imageProcessing.rotationEnable, bFalse);
      }catch (CmvBlueFOX3Exception& e) {
        throw CmvBlueFOX3Exception(_HERE_, "Unable to disable image rotation: " + e.what() );
      }  
    }
  }
}

void CMvbluefox3::SetImgMirror(int type)
{
  if (!this->bf_sets_->imageProcessing.mirrorModeGlobal.isValid())
    throw CmvBlueFOX3Exception(_HERE_, "Image mirror option is not available for this model.");  

  try{    
    WriteProperty(this->bf_sets_->imageProcessing.mirrorModeGlobal,type);
  }catch (CmvBlueFOX3Exception& e) {
    throw CmvBlueFOX3Exception(_HERE_, "Unable to mirror image: " + e.what() );
  }  
}

void CMvbluefox3::SetAoi(int &width, int &height, int &start_x, int &start_y) 
{
  if (this->params_.set_aoi)
  {
    if (!this->img_ctrl_->regionMode.isValid())
      throw CmvBlueFOX3Exception(_HERE_, "Area of Interest (AOI) option is not available for this model.");

    try{    
      WriteProperty(this->bf_sets_->cameraSetting.aoiWidth, width);
      WriteProperty(this->bf_sets_->cameraSetting.aoiHeight, height);
      WriteProperty(this->bf_sets_->cameraSetting.aoiStartX, start_x);
      WriteProperty(this->bf_sets_->cameraSetting.aoiStartY, start_y);
    }catch (CmvBlueFOX3Exception& e) {
      throw CmvBlueFOX3Exception(_HERE_, "Unable to set AOI: " + e.what() );
    }  
  }
}

void CMvbluefox3::SetHBinning(bool &h_bin_enbl, int &h_bin)
{
  if (h_bin_enbl)
  {
    if (!this->img_ctrl_->binningHorizontal.isValid())
      throw CmvBlueFOX3Exception(_HERE_, "Horizontal image binning option is not available for this model.");

    try{    
      WriteProperty(this->img_ctrl_->binningHorizontal, h_bin);
    }catch (CmvBlueFOX3Exception& e) {
      throw CmvBlueFOX3Exception(_HERE_, "Unable to Set horizontal image binning: " + e.what());
    }    
  }
}

void CMvbluefox3::SetVBinning(bool &v_bin_enbl, int &v_bin)
{
  if (v_bin_enbl)
  {
    if (!this->img_ctrl_->binningVertical.isValid())
      throw CmvBlueFOX3Exception(_HERE_, "Vertical image binning option is not available for this model.");

    try{    
      WriteProperty(this->img_ctrl_->binningVertical, v_bin);
    }catch (CmvBlueFOX3Exception& e) {
      throw CmvBlueFOX3Exception(_HERE_, "Unable to Set vertical image binning: " + e.what());
    }    
  }
}

void CMvbluefox3::SetAutoGainCtrl(bool &auto_gain, float &gain_db) 
{
  if (!this->anlg_ctrl_->gainAuto.isValid())
    throw CmvBlueFOX3Exception(_HERE_, "Analogic Auto Gain Control option is not available for this model.");

  try{    
    WriteProperty(this->anlg_ctrl_->gainAuto, auto_gain);
    if (!auto_gain) 
      WriteProperty(this->anlg_ctrl_->gain, gain_db);
    }catch (CmvBlueFOX3Exception& e) {
    throw CmvBlueFOX3Exception(_HERE_, "Unable to Set Analogic Auto Gain Control: " + e.what() );
  }
}

void CMvbluefox3::SetAutoExposCtrl(bool &auto_expose, int &auto_upperlimit, int &des_gray_val, int &expose_us)
{
  if (!this->ac_ctrl_->exposureAuto.isValid())
    throw CmvBlueFOX3Exception(_HERE_, "Auto Exposure Control option is not available for this model.");

  try{    
    if (auto_expose) 
    {
      WriteProperty(this->ac_ctrl_->exposureAuto, 1);
      WriteProperty(this->ac_ctrl_->mvExposureAutoUpperLimit, auto_upperlimit);
      WriteProperty(this->ac_ctrl_->mvExposureAutoAverageGrey, des_gray_val);
    }
    else 
    {
      WriteProperty(this->ac_ctrl_->exposureAuto, 0);
      WriteProperty(this->ac_ctrl_->exposureTime, expose_us);
    }
  }catch (CmvBlueFOX3Exception& e) {
    throw CmvBlueFOX3Exception(_HERE_, "Unable to Set Auto Exposure Control: " + e.what() );
  }
} 

void CMvbluefox3::SetAutoCtrl(bool &actrl_enbl, int &auto_ctrl_speedpeed, int &des_gray_val)
{
  if (actrl_enbl)
  {
    if (!this->ac_ctrl_->mvExposureAutoSpeed.isValid())
      throw CmvBlueFOX3Exception(_HERE_, "Auto Speed Control option is not available for this model.");  
    try{      
      bool agc = false, aec = false;
      ReadProperty(this->anlg_ctrl_->gainAuto, agc);
      ReadProperty(this->ac_ctrl_->exposureAuto, aec);
      if (agc || aec) 
      {
        WriteProperty(this->ac_ctrl_->mvExposureAutoSpeed, auto_ctrl_speedpeed);
        WriteProperty(this->ac_ctrl_->mvExposureAutoAverageGrey, des_gray_val);
      }
    }catch (CmvBlueFOX3Exception& e) {
      throw CmvBlueFOX3Exception(_HERE_, "Unable to Set Auto Exposure Control: "   + e.what() );  
    }    
  }
}

void CMvbluefox3::SetWhiteBlnce(bool &wbp, float &r_gain, float &b_gain)
{
  TImageDestinationPixelFormat pf = this->bf_sets_->imageDestination.pixelFormat.read();

  // Check it's a colour camera
  if ( pf != idpfMono8 && pf != idpfMono10 && pf != idpfMono12 && pf != idpfMono12Packed_V1 && pf != idpfMono12Packed_V2 && pf != idpfMono14 && pf != idpfMono16)
  {
    if (wbp)
    {
      if (!this->anlg_ctrl_->balanceWhiteAuto.isValid())
        throw CmvBlueFOX3Exception(_HERE_, "Auto White Balance option is not available for this model.");
      try{      
        WriteProperty(this->anlg_ctrl_->balanceWhiteAuto,2);
      }catch (CmvBlueFOX3Exception& e) {
        throw CmvBlueFOX3Exception(_HERE_, "Unable to Set Auto White Balance: "   + e.what() );  
      }   
    }
    else
    {
      if (this->anlg_ctrl_->balanceWhiteAuto.isValid())
        WriteProperty(this->anlg_ctrl_->balanceWhiteAuto,0);

      if (this->anlg_ctrl_->balanceRatioSelector.isValid())
      {
        WriteProperty(this->anlg_ctrl_->balanceRatioSelector,0);
        WriteProperty(this->anlg_ctrl_->balanceRatio,r_gain);
        WriteProperty(this->anlg_ctrl_->balanceRatioSelector,1);
        WriteProperty(this->anlg_ctrl_->balanceRatio,b_gain);  
      }
    }
  }
}

void CMvbluefox3::SetHdr(bool &hdr)
{
  if (hdr) 
  {
    if (!this->bf_sets_->cameraSetting.getHDRControl().isAvailable())
      throw CmvBlueFOX3Exception(_HERE_, "HDR option is not available for this model.");    

    try{      
      WriteProperty(this->bf_sets_->cameraSetting.getHDRControl().HDREnable, hdr);
      WriteProperty(this->bf_sets_->cameraSetting.getHDRControl().HDRMode, cHDRmFixed0);
    }catch (CmvBlueFOX3Exception& e) {
      throw CmvBlueFOX3Exception(_HERE_, "Unable to Set Auto White Balance: "   + e.what() );  
    }  
  }
}

void CMvbluefox3::SetDarkCFilter(int &dcfm)
{
  if (dcfm != dcfmOff && !this->bf_sets_->imageProcessing.darkCurrentFilterMode.isValid())
      throw CmvBlueFOX3Exception(_HERE_, "Dark filter option is not available for this model.");    
   
  try{  
 
    if (dcfm == dcfmCalibrateDarkCurrent) 
    {
      // Special case for calibrate mode
      // Set "OffsetAutoCalibration = Off"
      if (this->bf_sets_->cameraSetting.offsetAutoCalibration.isValid())
        WriteProperty(this->bf_sets_->cameraSetting.offsetAutoCalibration, aocOff);
      // Set filter mode = calibrate
      WriteProperty(this->bf_sets_->imageProcessing.darkCurrentFilterMode, dcfmCalibrateDarkCurrent);
      // Read image count, and request some more images
      int img_cnt = this->bf_sets_->imageProcessing.darkCurrentFilterCalibrationImageCount.read();
      RequestMultiple(img_cnt);
      // Then turn on immediately
      WriteProperty(this->bf_sets_->imageProcessing.darkCurrentFilterMode, dcfmOn);
      if (this->bf_sets_->cameraSetting.offsetAutoCalibration.isValid())
        WriteProperty(this->bf_sets_->cameraSetting.offsetAutoCalibration, aocOn);
    } else {
      WriteAndReadProperty(this->bf_sets_->imageProcessing.darkCurrentFilterMode, dcfm);
    }  
  }catch (CmvBlueFOX3Exception& e) {
    throw CmvBlueFOX3Exception(_HERE_, "Unable to Set Auto White Balance: "   + e.what() );  
  }  
}

void CMvbluefox3::SetTimeoutMS(int &timeout_ms) 
{ 
  WriteProperty(this->bf_sets_->cameraSetting.imageRequestTimeout_ms,timeout_ms); 
  this->timeout_ms_ = timeout_ms;
}

int CMvbluefox3::GetTimeoutMS(void)
{ 
  int val;
  ReadProperty(this->bf_sets_->cameraSetting.imageRequestTimeout_ms,val);
  return val;
}

void CMvbluefox3::SetPixelClock(int &pclk_khz) 
{
  if (this->dev_ctrl_->mvDeviceClockFrequency.isValid())
  {
    try{      
      WriteProperty(this->dev_ctrl_->mvDeviceClockFrequency,pclk_khz);
      int fps = PixelClockToFrameRate(pclk_khz, params_.width, params_.height, params_.expose_us);
      SetFrameRate(fps);
      this->params_.frame_rate=fps;
    }catch (CmvBlueFOX3Exception& e) {
      throw CmvBlueFOX3Exception(_HERE_, "Unable to set pixel clock: " + e.what() );  
    }     
  }
}

void CMvbluefox3::SetFrameRate(int &rate) 
{
  try{      
    WriteProperty(this->ac_ctrl_->mvAcquisitionFrameRateLimitMode,0);
    WriteProperty(this->ac_ctrl_->acquisitionFrameRateEnable, bTrue);
    WriteProperty(this->ac_ctrl_->acquisitionFrameRate, rate);
  }catch (CmvBlueFOX3Exception& e) {
    throw CmvBlueFOX3Exception(_HERE_, "Unable to set frame rate: " + e.what() );  
  }        
}

int CMvbluefox3::PixelClockToFrameRate(int &pclk_khz, int &width, int &height, int &expose_us) 
{
  static const double kTriggerPulseWidthUs = 200;
  double frame_time_us = ((double)width + 61) * ((double)height + 45) / (double)pclk_khz * 1e3;
  return (int)(1e6 / (frame_time_us + (double)expose_us + kTriggerPulseWidthUs));
}

void CMvbluefox3::SetGamma(bool &enbl)
{
  if (!this->lut_ctrl_->LUTEnable.isValid())
    throw CmvBlueFOX3Exception(_HERE_, "LUT option is not available for this model.");  
  try{   
    if (enbl)
    {
      WriteProperty(this->lut_ctrl_->LUTEnable,bTrue);   
    }
    else
      WriteProperty(this->lut_ctrl_->LUTEnable,bFalse);   
  }catch (CmvBlueFOX3Exception& e) {
    throw CmvBlueFOX3Exception(_HERE_, "Unable to set gamma: " + e.what() );  
  }  
}

void CMvbluefox3::SetBlackLevel(bool &auto_blck_level, int &blck_level)
{
  if (auto_blck_level)
  {
    if (!this->anlg_ctrl_->blackLevelAuto.isValid())
      throw CmvBlueFOX3Exception(_HERE_, "Auto black level option is not available for this model."); 

    try{     
      WriteProperty(this->anlg_ctrl_->blackLevelAuto,2);
      WriteProperty(this->anlg_ctrl_->blackLevel,0);
    }catch (CmvBlueFOX3Exception& e) {
      throw CmvBlueFOX3Exception(_HERE_, "Unable to set gamma: " + e.what() );  
    }  
  }
  else
  {
    if (!this->anlg_ctrl_->blackLevel.isValid())
      throw CmvBlueFOX3Exception(_HERE_, "Black level option is not available for this model."); 
    try{     
      WriteProperty(this->anlg_ctrl_->blackLevel,blck_level);
      WriteProperty(this->anlg_ctrl_->blackLevelAuto,0);
    }catch (CmvBlueFOX3Exception& e) {
      throw CmvBlueFOX3Exception(_HERE_, "Unable to set gamma: " + e.what() );  
    }  
  }
}

void CMvbluefox3::ColorTwistFilter(bool &ct_enbl)
{
  // TODO: Color twist filter (ImageProcessing/ColorTwist) with setSaturation().

  if (ct_enbl)
  {
    if (!this->bf_sets_->imageProcessing.colorTwistEnable.isValid())
      throw CmvBlueFOX3Exception(_HERE_, "Color twist filter option is not available for this model."); 

    try{     
      WriteProperty(this->bf_sets_->imageProcessing.colorTwistEnable,bTrue);
      this->bf_sets_->imageProcessing.setSaturation(1);
      WriteProperty(this->bf_sets_->imageProcessing.colorTwistInputCorrectionMatrixEnable,bTrue);
      WriteProperty(this->bf_sets_->imageProcessing.colorTwistOutputCorrectionMatrixEnable,bTrue);
      WriteProperty(this->bf_sets_->imageProcessing.colorTwistInputCorrectionMatrixMode,cticmmDeviceSpecific);
      WriteProperty(this->bf_sets_->imageProcessing.colorTwistOutputCorrectionMatrixMode,ctocmmXYZTosRGB_D50);
      WriteProperty(this->bf_sets_->imageProcessing.colorTwistInputCorrectionMatrixEnable,bFalse);
      WriteProperty(this->bf_sets_->imageProcessing.colorTwistOutputCorrectionMatrixEnable,bFalse);
    }catch (CmvBlueFOX3Exception& e) {
      throw CmvBlueFOX3Exception(_HERE_, "Unable to set color twist filter: " + e.what() );  
    }  
  }  
  else
  {
    if (this->bf_sets_->imageProcessing.colorTwistEnable.isValid())
    {
      try{     
        WriteProperty(this->bf_sets_->imageProcessing.colorTwistEnable,bFalse);
        WriteProperty(this->bf_sets_->imageProcessing.colorTwistInputCorrectionMatrixEnable,bFalse);
        WriteProperty(this->bf_sets_->imageProcessing.colorTwistOutputCorrectionMatrixEnable,bFalse);
      }catch (CmvBlueFOX3Exception& e) {
        throw CmvBlueFOX3Exception(_HERE_, "Unable to set color twist filter: " + e.what() );  
      }        
    }
  }
}

} // End of CMvbluefox3 namespace